//	2000/07/22 Ogasawara Hiroyuki

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<math.h>
#include	"_dirent.h"
#include	<time.h>

#define		stat	__linux_stat
#include	"/usr/local/zxksrc/include/asm/stat.h"
#undef		stat

enum {
	SHORT_FORMAT,
	LONG_FORMAT,
};

#define		OPTION_ALL		(1<<0)
#define		OPTION_INODE		(1<<1)
#define		OPTION_INDICATOR	(1<<2)

static int	ListFormat= SHORT_FORMAT;
static int	ListCount= 0;
static int	ListOption= OPTION_INDICATOR;

#define		S_COLUMNS		72
#define		S_FNLEN			15
#define		S_FNLIST		(S_COLUMNS/S_FNLEN)

static void
_filestat( char *file )
{
	struct stat	fbuf;
	struct tm	*tmp;
	unsigned char	option= '\0',
			type= '-';
	char		modestr[10]; // trwxrwxrwx
	char		linkbuf[256];
//	struct __linux_stat	fbuf;
	if( lstat( file, (struct stat*)&fbuf ) < 0 )
		return;
	tmp= localtime( &fbuf.st_mtime );
	if( fbuf.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH) ){
		option= '*';
	}
	if( S_ISDIR(fbuf.st_mode) ){
		option= '/';
		type= 'd';
	}
	*linkbuf= '\0';
	if( S_ISLNK(fbuf.st_mode) ){
		if( ListFormat == LONG_FORMAT ){
			int	len= readlink( file, linkbuf, 255 );
			struct stat	lfbuf;
			option= ' ';
			if( stat( linkbuf, (struct stat*)&lfbuf ) >= 0 ){
				if( S_ISDIR(lfbuf.st_mode) ){
					linkbuf[len++]= '/';
				}else if( lfbuf.st_mode &
						(S_IXUSR|S_IXGRP|S_IXOTH) ){
					linkbuf[len++]= '*';
				}
			}
			linkbuf[len]= '\0';
		}else{
			option= '@';
		}
		type= 'l';
	}
	if( S_ISBLK(fbuf.st_mode) ){
		type= 'b';
	}
	if( S_ISCHR(fbuf.st_mode) ){
		type= 'c';
	}
	if( !(ListOption & OPTION_INDICATOR) ){
		option= '\0';
	}
	{
		int	i;
		int	md= fbuf.st_mode;
		char	*mdp= modestr;
		for( i= 0 ; i< 9 ; i++, md<<= 1 ){
			*mdp++= (md & 0400) ? ((char*)"rwxrwxrwx")[i] : '-';
		}
		*mdp= '\0';
	}
	switch( ListFormat ){
	case LONG_FORMAT:
		printf(
"%c%s%3d%6d%6d%10d %04d-%02d-%02d %02d:%02d:%02d %s%c",
			type,
			modestr,
			fbuf.st_nlink,
			fbuf.st_uid,
			fbuf.st_gid,
			fbuf.st_size,
			tmp->tm_year+1900,
			tmp->tm_mon+1,
			tmp->tm_mday,
			tmp->tm_hour,
			tmp->tm_min,
			tmp->tm_sec,
			file,
			option
		);
		if( *linkbuf ){
			fputs( "-> ", stdout );
			puts( linkbuf );
		}else{
			putchar( '\n' );
		}
		break;
#if 0
	case LONG_FORMAT:
		printf(
"%c%s%8d %02d-%02d %02d:%02d:%02d %s%c",
			type,
			modestr,
			fbuf.st_size,
			tmp->tm_mon+1,
			tmp->tm_mday,
			tmp->tm_hour,
			tmp->tm_min,
			tmp->tm_sec,
			file,
			option
		);
		putchar( '\n' );
		break;
#endif
	case SHORT_FORMAT:
		fputs( file, stdout );
		if( option )
			putchar( option );
		if( ListCount < (S_FNLIST-1) ){
			int	i= S_FNLEN-strlen( file );
			if( option )
				i--;
			for(; i-- > 0 ; putchar( ' ' ) );
			ListCount++;
		}else{
			putchar( '\n' );
			ListCount= 0;
		}
		break;
	}
}

static void
_lsdir( char *path )
{
	DIR	*dirp;
	struct stat	fbuf;
	stat( path, &fbuf );
	if( !(S_ISDIR(fbuf.st_mode)) ){
		_filestat( path );
		return;
	}
	if( dirp= opendir( path ) ){
		struct _dirent	*dirdata;
		for(; dirdata= _readdir( dirp ) ;){
			if( !(ListOption & OPTION_ALL) ){
				if( *dirdata->d_name == '.' )
					continue;
			}
			if( *path == '.' && !path[1] ){
				_filestat( dirdata->d_name );
			}else{
				char	_maxpath[strlen(path)+strlen(dirdata->d_name)+4];
				strcpy( _maxpath, path );
				strcat( _maxpath, "/" );
				strcat( _maxpath, dirdata->d_name );
				_filestat( _maxpath );
			}
		}
		closedir( dirp );
	}
}

int
main( int argc, char **argv )
{
	if( argc >= 2 ){
		int	i= 1;
		int	list= 0;
		for(; i < argc ; i++ ){
			if( argv[i][0] == '-' ){
				char	*p= &argv[i][1];
				for(; *p ; p++ ){
					switch( *p ){
					case 'l':
						ListFormat= LONG_FORMAT;
						break;
					case 'a':
						ListOption|= OPTION_ALL;
						break;
					case 'i':
						ListOption|= OPTION_INODE;
						break;
					case 'F':
						ListOption|= OPTION_INDICATOR;
						ListOption&=~OPTION_INDICATOR;
						break;
					case 'C':
						ListFormat= SHORT_FORMAT;
						break;
					}
				}
			}else
				list++;
		}
		if( list ){
			for(; argc >= 2 ; argc--, argv++ ){
				if( argv[1][0] != '-' )
					_lsdir( argv[1] );
			}
		}else{
			_lsdir( "." );
		}
	}else{
		_lsdir( "." );
	}

	if( ListFormat == SHORT_FORMAT && ListCount )
		putchar( '\n' );

	return	0;
}

